/*----------------------------------------------------------------------*/
/* Copyright (c) 1987							*/
/* by CompuServe Inc., Columbus, Ohio.  All Rights Reserved		*/
/*----------------------------------------------------------------------*/

#include <stdio.h>
#include "GIFbase.h"
#include "GIFWrite.h"
#include "compress.h"

static short Write_Color_Map (
/*
 * Function:
 *	Write the specified color map.
 *
 * Inputs:
 *	(see above)
 *
 * Returns:
 *	0 = OK, else error code
 */
    short (*write_byte)(unsigned char),
    short cr,
    short bp,
    short normalize,
    struct ColorEntry color_map[] );


short Create_GIF (
/*
 * Function:
 *	Create a 1-image GIF image-set.  The size of the screen will be the
 *	same as the image.
 *
 * Inputs:
 */
    short (*read_pixel)(void),		/* ptr to caller's pixel reader */
    short (*write_byte)(unsigned char),	/* ptr to caller's byte writer */
    short width,			/* width of image in pixels */
    short height,			/* height of image in pixels */
    short cr,				/* bits of color resolution (1..8) */
    short interlaced,			/* interlaced flag */
    short bp,				/* bits per pixel (1..8) */
    short normalize,			/* normalize the color map */
    struct ColorEntry color_map[] )	/* ptr to color map */
/*
 * Returns:
 *	0 = OK, else error code
 */
    {
    short status;		/* return status */

    status = Write_Screen_Desc (write_byte,
    				width, height, cr, 0,
    				bp, normalize, color_map);
    if (status != 0) return status;

    status = Write_Image_Desc (write_byte, 0, 0, width, height,
			       0, interlaced, 0, 0, NULL);

    if (status != 0) return status;

    status = Compress_Data (bp, read_pixel, write_byte);
    if (status != 0) return status;

    return (*write_byte)(';');
    }

/*$page*/

short Write_Screen_Desc (
/*
 * Function:
 *	Write the GIF signature, the screen description, and the optional
 *	color map.
 *
 * Inputs:
 */
    short (*write_byte)(unsigned char),	/* ptr to caller's byte writer */
    short width,			/* width of the screen in pixels */
    short height,			/* height of the screen in pixels */
    short cr,				/* bits of color resolution (1..8) */
    short fill_color,			/* pixel value used to fill the background */
    short bp,				/* bits per pixel (1..8) */
    short normalize,			/* normalize the color map */
    struct ColorEntry color_map[] )	/* ptr to the optional color map */
/*
 * Returns:
 *	0 = OK, else error code
 */
    {
    short status;
    short i;
    unsigned char buffer[16];

    for (i = 0; i < 6; i++)
	if ((status = (*write_byte)(GIFsignature[i])) != 0)
	    return status;

    buffer[0] = (unsigned char)(width & 0xFF);
    buffer[1] = (unsigned char)(width >> 8);
    buffer[2] = (unsigned char)(height & 0xFF);
    buffer[3] = (unsigned char)(height >> 8);

    if (color_map == NULL)
	buffer[4] = (unsigned char)(((cr - 1) & 0x07) << 4);
    else
	buffer[4] = (unsigned char)(0x80 | ((cr - 1) << 4) | ((bp - 1) & 0x0F));

    buffer[5] = (unsigned char)(fill_color);
    buffer[6] = 0;			/* reserved */

    for (i = 0; i < 7; i++)
	if ((status = (*write_byte)(buffer[i])) != 0)
	    return status;

    if (color_map != NULL)
	return Write_Color_Map (write_byte, cr, bp, normalize, color_map);
    else
	return 0;
    }

short Write_Image_Desc (
/*
 * Function:
 *	Write the image description and optional local color map.
 *
 * Inputs:
 */
    short (*write_byte)(unsigned char),	/* ptr to caller's byte writer */
    short left_edge,			/* (left_edge, top_edge) is the upper left */
    short top_edge,			/* position in pixels of the image */
    short width,			/* width of the image in pixels */
    short height,			/* height of the image in pixels */
    short cr,				/* bits of color resolution (1..8) */
    short interlaced,			/* 1 = interlace the image */
    short bp,				/* bits per pixel for this image */
    short normalize,			/* normalize the color map */
    struct ColorEntry color_map[] )	/* ptr to local color map */
/*
 * Returns:
 *	0 = OK, else error code
 */
    {
    short status;
    short i;
    unsigned char buffer[16];

    interlaced &= 1;			/* low bit only */

    buffer[0] = ',';
    buffer[1] = (unsigned char)(left_edge & 0xFF);
    buffer[2] = (unsigned char)(left_edge >> 8);
    buffer[3] = (unsigned char)(top_edge & 0xFF);
    buffer[4] = (unsigned char)(top_edge >> 8);
    buffer[5] = (unsigned char)(width & 0xFF);
    buffer[6] = (unsigned char)(width >> 8);
    buffer[7] = (unsigned char)(height & 0xFF);
    buffer[8] = (unsigned char)(height >> 8);

    if (color_map == NULL)
	buffer[9] = (unsigned char)(interlaced << 6);
    else
	buffer[9] = (unsigned char)
			(0x80 | (interlaced << 6) | ((bp - 1) & 0x0F));

    for (i = 0; i < 10; i++)
	if ((status = (*write_byte)(buffer[i])) != 0)
	    return status;

    if (color_map != NULL)
	Write_Color_Map (write_byte, cr, bp, normalize, color_map);
    else
	return 0;
    }

/*$page*/

static short Write_Color_Map (
/*
 * Function:
 *	Write the specified color map.
 *
 * Inputs:
 *	(see above)
 *
 * Returns:
 *	0 = OK, else error code
 */
    short (*write_byte)(unsigned char),
    short cr,
    short bp,
    short normalize,
    struct ColorEntry color_map[] )
    {
    short num_colors;
    short n, i, color;
    short status;

    num_colors = 1 << bp;
    n = (1 << cr) - 1;

    for (i = 0; i < num_colors; i++)
	{
	color = color_map[i].red;
	if (normalize) color = color*255/n;
	status = (*write_byte)((unsigned char)(color));
	if (status != 0) return status;

	color = color_map[i].green;
	if (normalize) color = color*255/n;
	status = (*write_byte)((unsigned char)(color));
	if (status != 0) return status;

	color = color_map[i].blue;
	if (normalize) color = color*255/n;
	status = (*write_byte)((unsigned char)(color));
	if (status != 0) return status;
	}

    return 0;
    }
